@@ -15,22 +15,14 @@ a single value.
 Notably, the set of steps available at any point may depend on the
 execution to date.
 """
-
+import collections
 import inspect
+from collections.abc import Iterable, Sequence
 from copy import copy
 from functools import lru_cache
 from io import StringIO
-from typing import (
-    Any,
-    Callable,
-    Dict,
-    Iterable,
-    List,
-    Optional,
-    Sequence,
-    Union,
-    overload,
-)
+from time import perf_counter
+from typing import Any, Callable, ClassVar, Optional, Union, overload
 from unittest import TestCase
 
 import attr
@@ -42,11 +34,15 @@ from hypothesis._settings import (
     note_deprecation,
     settings as Settings,
 )
-from hypothesis.control import current_build_context
+from hypothesis.control import _current_build_context, current_build_context
 from hypothesis.core import TestFunc, given
 from hypothesis.errors import InvalidArgument, InvalidDefinition
+from hypothesis.internal.compat import add_note
 from hypothesis.internal.conjecture import utils as cu
+from hypothesis.internal.conjecture.engine import BUFFER_SIZE_IR
+from hypothesis.internal.conjecture.junkdrawer import gc_cumulative_time
 from hypothesis.internal.healthcheck import fail_health_check
+from hypothesis.internal.observability import TESTCASE_CALLBACKS
 from hypothesis.internal.reflection import (
     function_digest,
     get_pretty_function_description,
@@ -87,21 +83,19 @@ class TestCaseProperty:  # pragma: no cover
         raise AttributeError("Cannot delete TestCase")
 
 
-def run_state_machine_as_test(state_machine_factory, *, settings=None):
-    """Run a state machine definition as a test, either silently doing nothing
-    or printing a minimal breaking program and raising an exception.
-
-    state_machine_factory is anything which returns an instance of
-    RuleBasedStateMachine when called with no arguments - it can be a class or a
-    function. settings will be used to control the execution of the test.
-    """
+def get_state_machine_test(state_machine_factory, *, settings=None, _min_steps=0):
     if settings is None:
         try:
             settings = state_machine_factory.TestCase.settings
             check_type(Settings, settings, "state_machine_factory.TestCase.settings")
         except AttributeError:
-            settings = Settings(deadline=None, suppress_health_check=HealthCheck.all())
+            settings = Settings(deadline=None, suppress_health_check=list(HealthCheck))
     check_type(Settings, settings, "settings")
+    check_type(int, _min_steps, "_min_steps")
+    if _min_steps < 0:
+        # Because settings can vary via e.g. profiles, settings.stateful_step_count
+        # overrides this argument and we don't bother cross-validating.
+        raise InvalidArgument(f"_min_steps={_min_steps} must be non-negative.")
 
     @settings
     @given(st.data())
@@ -110,14 +104,22 @@ def run_state_machine_as_test(state_machine_factory, *, settings=None):
         machine = factory()
         check_type(RuleBasedStateMachine, machine, "state_machine_factory()")
         cd.hypothesis_runner = machine
+        machine._observability_predicates = cd._observability_predicates  # alias
 
         print_steps = (
             current_build_context().is_final or current_verbosity() >= Verbosity.debug
         )
-        try:
+        cd._stateful_repr_parts = []
+
+        def output(s):
             if print_steps:
-                report(f"state = {machine.__class__.__name__}()")
-            machine.check_invariants(settings)
+                report(s)
+            if TESTCASE_CALLBACKS:
+                cd._stateful_repr_parts.append(s)
+
+        try:
+            output(f"state = {machine.__class__.__name__}()")
+            machine.check_invariants(settings, output, cd._stateful_run_times)
             max_steps = settings.stateful_step_count
             steps_run = 0
 
@@ -129,24 +131,20 @@ def run_state_machine_as_test(state_machine_factory, *, settings=None):
                 # 2 ** -16 during normal operation but force a stop when we've
                 # generated enough steps.
                 cd.start_example(STATE_MACHINE_RUN_LABEL)
-                if steps_run == 0:
-                    cd.draw_bits(16, forced=1)
-                elif steps_run >= max_steps:
-                    cd.draw_bits(16, forced=0)
+                must_stop = None
+                if steps_run >= max_steps:
+                    must_stop = True
+                elif steps_run <= _min_steps:
+                    must_stop = False
+                elif cd.length_ir > (0.8 * BUFFER_SIZE_IR):
+                    # Better to stop after fewer steps, than always overrun and retry.
+                    # See https://github.com/HypothesisWorks/hypothesis/issues/3618
+                    must_stop = True
+
+                start_draw = perf_counter()
+                start_gc = gc_cumulative_time()
+                if cd.draw_boolean(p=2**-16, forced=must_stop):
                     break
-                else:
-                    # All we really care about is whether this value is zero
-                    # or non-zero, so if it's > 1 we discard it and insert a
-                    # replacement value after
-                    cd.start_example(SHOULD_CONTINUE_LABEL)
-                    should_continue_value = cd.draw_bits(16)
-                    if should_continue_value > 1:
-                        cd.stop_example(discard=True)
-                        cd.draw_bits(16, forced=int(bool(should_continue_value)))
-                    else:
-                        cd.stop_example()
-                        if should_continue_value == 0:
-                            break
                 steps_run += 1
 
                 # Choose a rule to run, preferring an initialize rule if there are
@@ -160,12 +158,16 @@ def run_state_machine_as_test(state_machine_factory, *, settings=None):
                     machine._initialize_rules_to_run.remove(rule)
                 else:
                     rule, data = cd.draw(machine._rules_strategy)
+                draw_label = f"generate:rule:{rule.function.__name__}"
+                cd.draw_times.setdefault(draw_label, 0.0)
+                in_gctime = gc_cumulative_time() - start_gc
+                cd.draw_times[draw_label] += perf_counter() - start_draw - in_gctime
 
                 # Pretty-print the values this rule was called with *before* calling
                 # _add_result_to_targets, to avoid printing arguments which are also
                 # a return value using the variable name they are assigned to.
                 # See https://github.com/HypothesisWorks/hypothesis/issues/2341
-                if print_steps:
+                if print_steps or TESTCASE_CALLBACKS:
                     data_to_print = {
                         k: machine._pretty_print(v) for k, v in data.items()
                     }
@@ -177,7 +179,18 @@ def run_state_machine_as_test(state_machine_factory, *, settings=None):
                     for k, v in list(data.items()):
                         if isinstance(v, VarReference):
                             data[k] = machine.names_to_values[v.name]
+                        elif isinstance(v, list) and all(
+                            isinstance(item, VarReference) for item in v
+                        ):
+                            data[k] = [machine.names_to_values[item.name] for item in v]
+
+                    label = f"execute:rule:{rule.function.__name__}"
+                    start = perf_counter()
+                    start_gc = gc_cumulative_time()
                     result = rule.function(machine, **data)
+                    in_gctime = gc_cumulative_time() - start_gc
+                    cd._stateful_run_times[label] += perf_counter() - start - in_gctime
+
                     if rule.targets:
                         if isinstance(result, MultipleResults):
                             for single_result in result.values:
@@ -194,16 +207,15 @@ def run_state_machine_as_test(state_machine_factory, *, settings=None):
                             HealthCheck.return_value,
                         )
                 finally:
-                    if print_steps:
+                    if print_steps or TESTCASE_CALLBACKS:
                         # 'result' is only used if the step has target bundles.
                         # If it does, and the result is a 'MultipleResult',
                         # then 'print_step' prints a multi-variable assignment.
-                        machine._print_step(rule, data_to_print, result)
-                machine.check_invariants(settings)
+                        output(machine._repr_step(rule, data_to_print, result))
+                machine.check_invariants(settings, output, cd._stateful_run_times)
                 cd.stop_example()
         finally:
-            if print_steps:
-                report("state.teardown()")
+            output("state.teardown()")
             machine.teardown()
 
     # Use a machine digest to identify stateful tests in the example database
@@ -218,19 +230,31 @@ def run_state_machine_as_test(state_machine_factory, *, settings=None):
         state_machine_factory, "_hypothesis_internal_use_reproduce_failure", None
     )
     run_state_machine._hypothesis_internal_print_given_args = False
+    return run_state_machine
 
-    run_state_machine(state_machine_factory)
+
+def run_state_machine_as_test(state_machine_factory, *, settings=None, _min_steps=0):
+    """Run a state machine definition as a test, either silently doing nothing
+    or printing a minimal breaking program and raising an exception.
+
+    state_machine_factory is anything which returns an instance of
+    RuleBasedStateMachine when called with no arguments - it can be a class or a
+    function. settings will be used to control the execution of the test.
+    """
+    state_machine_test = get_state_machine_test(
+        state_machine_factory, settings=settings, _min_steps=_min_steps
+    )
+    state_machine_test(state_machine_factory)
 
 
 class StateMachineMeta(type):
     def __setattr__(cls, name, value):
         if name == "settings" and isinstance(value, Settings):
+            descr = f"settings({value.show_changed()})"
             raise AttributeError(
-                (
-                    "Assigning {cls}.settings = {value} does nothing. Assign "
-                    "to {cls}.TestCase.settings, or use @{value} as a decorator "
-                    "on the {cls} class."
-                ).format(cls=cls.__name__, value=value)
+                f"Assigning {cls.__name__}.settings = {descr} does nothing. Assign "
+                f"to {cls.__name__}.TestCase.settings, or use @{descr} as a decorator "
+                f"on the {cls.__name__} class."
             )
         return super().__setattr__(name, value)
 
@@ -238,31 +262,48 @@ class StateMachineMeta(type):
 class RuleBasedStateMachine(metaclass=StateMachineMeta):
     """A RuleBasedStateMachine gives you a structured way to define state machines.
 
-    The idea is that a state machine carries a bunch of types of data
-    divided into Bundles, and has a set of rules which may read data
-    from bundles (or just from normal strategies) and push data onto
-    bundles. At any given point a random applicable rule will be
-    executed.
+    The idea is that a state machine carries the system under test and some supporting
+    data. This data can be stored in instance variables or
+    divided into Bundles. The state machine has a set of rules which may read data
+    from bundles (or just from normal strategies), push data onto
+    bundles, change the state of the machine, or verify properties.
+    At any given point a random applicable rule will be executed.
     """
 
-    _rules_per_class: Dict[type, List[classmethod]] = {}
-    _invariants_per_class: Dict[type, List[classmethod]] = {}
-    _initializers_per_class: Dict[type, List[classmethod]] = {}
+    _rules_per_class: ClassVar[dict[type, list[classmethod]]] = {}
+    _invariants_per_class: ClassVar[dict[type, list[classmethod]]] = {}
+    _initializers_per_class: ClassVar[dict[type, list[classmethod]]] = {}
 
-    def __init__(self):
+    def __init__(self) -> None:
         if not self.rules():
             raise InvalidDefinition(f"Type {type(self).__name__} defines no rules")
-        self.bundles: Dict[str, list] = {}
-        self.name_counter = 1
-        self.names_to_values: Dict[str, Any] = {}
+        self.bundles: dict[str, list] = {}
+        self.names_counters: collections.Counter = collections.Counter()
+        self.names_list: list[str] = []
+        self.names_to_values: dict[str, Any] = {}
         self.__stream = StringIO()
-        self.__printer = RepresentationPrinter(self.__stream)
+        self.__printer = RepresentationPrinter(
+            self.__stream, context=_current_build_context.value
+        )
         self._initialize_rules_to_run = copy(self.initialize_rules())
         self._rules_strategy = RuleStrategy(self)
 
+        if isinstance(s := vars(type(self)).get("settings"), Settings):
+            tname = type(self).__name__
+            descr = f"settings({s.show_changed()})"
+            raise InvalidDefinition(
+                f"Assigning settings = {descr} as a class attribute does nothing. "
+                f"Assign to {tname}.TestCase.settings, or use @{descr} as a decorator "
+                f"on the {tname} class."
+            )
+
     def _pretty_print(self, value):
         if isinstance(value, VarReference):
             return value.name
+        elif isinstance(value, list) and all(
+            isinstance(item, VarReference) for item in value
+        ):
+            return "[" + ", ".join([item.name for item in value]) + "]"
         self.__stream.seek(0)
         self.__stream.truncate(0)
         self.__printer.output_width = 0
@@ -275,15 +316,16 @@ class RuleBasedStateMachine(metaclass=StateMachineMeta):
     def __repr__(self):
         return f"{type(self).__name__}({nicerepr(self.bundles)})"
 
-    def _new_name(self):
-        result = f"v{self.name_counter}"
-        self.name_counter += 1
+    def _new_name(self, target):
+        result = f"{target}_{self.names_counters[target]}"
+        self.names_counters[target] += 1
+        self.names_list.append(result)
         return result
 
     def _last_names(self, n):
-        assert self.name_counter > n
-        count = self.name_counter
-        return [f"v{i}" for i in range(count - n, count)]
+        len_ = len(self.names_list)
+        assert len_ >= n
+        return self.names_list[len_ - n :]
 
     def bundle(self, name):
         return self.bundles.setdefault(name, [])
@@ -331,47 +373,48 @@ class RuleBasedStateMachine(metaclass=StateMachineMeta):
         cls._invariants_per_class[cls] = target
         return cls._invariants_per_class[cls]
 
-    def _print_step(self, rule, data, result):
-        self.step_count = getattr(self, "step_count", 0) + 1
+    def _repr_step(self, rule, data, result):
         output_assignment = ""
         if rule.targets:
             if isinstance(result, MultipleResults):
                 if len(result.values) == 1:
                     output_assignment = f"({self._last_names(1)[0]},) = "
                 elif result.values:
-                    output_names = self._last_names(len(result.values))
+                    number_of_last_names = len(rule.targets) * len(result.values)
+                    output_names = self._last_names(number_of_last_names)
                     output_assignment = ", ".join(output_names) + " = "
             else:
                 output_assignment = self._last_names(1)[0] + " = "
-        report(
-            "{}state.{}({})".format(
-                output_assignment,
-                rule.function.__name__,
-                ", ".join("%s=%s" % kv for kv in data.items()),
-            )
-        )
+        args = ", ".join("%s=%s" % kv for kv in data.items())
+        return f"{output_assignment}state.{rule.function.__name__}({args})"
 
     def _add_result_to_targets(self, targets, result):
-        name = self._new_name()
-        self.__printer.singleton_pprinters.setdefault(
-            id(result), lambda obj, p, cycle: p.text(name)
-        )
-        self.names_to_values[name] = result
         for target in targets:
+            name = self._new_name(target)
+
+            def printer(obj, p, cycle, name=name):
+                return p.text(name)
+
+            self.__printer.singleton_pprinters.setdefault(id(result), printer)
+            self.names_to_values[name] = result
             self.bundles.setdefault(target, []).append(VarReference(name))
 
-    def check_invariants(self, settings):
+    def check_invariants(self, settings, output, runtimes):
         for invar in self.invariants():
             if self._initialize_rules_to_run and not invar.check_during_init:
                 continue
             if not all(precond(self) for precond in invar.preconditions):
                 continue
+            name = invar.function.__name__
             if (
                 current_build_context().is_final
                 or settings.verbosity >= Verbosity.debug
+                or TESTCASE_CALLBACKS
             ):
-                report(f"state.{invar.function.__name__}()")
+                output(f"state.{name}()")
+            start = perf_counter()
             result = invar.function(self)
+            runtimes[f"execute:invariant:{name}"] += perf_counter() - start
             if result is not None:
                 fail_health_check(
                     settings,
@@ -391,22 +434,23 @@ class RuleBasedStateMachine(metaclass=StateMachineMeta):
     TestCase = TestCaseProperty()
 
     @classmethod
-    @lru_cache()
+    @lru_cache
     def _to_test_case(cls):
         class StateMachineTestCase(TestCase):
-            settings = Settings(deadline=None, suppress_health_check=HealthCheck.all())
+            settings = Settings(deadline=None, suppress_health_check=list(HealthCheck))
 
             def runTest(self):
-                run_state_machine_as_test(cls)
+                run_state_machine_as_test(cls, settings=self.settings)
 
             runTest.is_hypothesis_test = True
+            runTest._hypothesis_state_machine_class = cls
 
         StateMachineTestCase.__name__ = cls.__name__ + ".TestCase"
         StateMachineTestCase.__qualname__ = cls.__qualname__ + ".TestCase"
         return StateMachineTestCase
 
 
-@attr.s()
+@attr.s(repr=False)
 class Rule:
     targets = attr.ib()
     function = attr.ib(repr=get_pretty_function_description)
@@ -415,25 +459,28 @@ class Rule:
     bundles = attr.ib(init=False)
 
     def __attrs_post_init__(self):
-        arguments = {}
+        self.arguments_strategies = {}
         bundles = []
         for k, v in sorted(self.arguments.items()):
             assert not isinstance(v, BundleReferenceStrategy)
             if isinstance(v, Bundle):
                 bundles.append(v)
                 consume = isinstance(v, BundleConsumer)
-                arguments[k] = BundleReferenceStrategy(v.name, consume)
-            else:
-                arguments[k] = v
+                v = BundleReferenceStrategy(v.name, consume=consume)
+            self.arguments_strategies[k] = v
         self.bundles = tuple(bundles)
-        self.arguments_strategy = st.fixed_dictionaries(arguments)
+
+    def __repr__(self) -> str:
+        rep = get_pretty_function_description
+        bits = [f"{k}={rep(v)}" for k, v in attr.asdict(self).items() if v]
+        return f"{self.__class__.__name__}({', '.join(bits)})"
 
 
 self_strategy = st.runner()
 
 
 class BundleReferenceStrategy(SearchStrategy):
-    def __init__(self, name, consume=False):
+    def __init__(self, name: str, *, consume: bool = False):
         self.name = name
         self.consume = consume
 
@@ -441,21 +488,45 @@ class BundleReferenceStrategy(SearchStrategy):
         machine = data.draw(self_strategy)
         bundle = machine.bundle(self.name)
         if not bundle:
-            data.mark_invalid()
+            data.mark_invalid(f"Cannot draw from empty bundle {self.name!r}")
         # Shrink towards the right rather than the left. This makes it easier
         # to delete data generated earlier, as when the error is towards the
         # end there can be a lot of hard to remove padding.
-        position = cu.integer_range(data, 0, len(bundle) - 1, center=len(bundle))
+        position = data.draw_integer(0, len(bundle) - 1, shrink_towards=len(bundle))
         if self.consume:
-            return bundle.pop(position)
+            return bundle.pop(position)  # pragma: no cover  # coverage is flaky here
         else:
             return bundle[position]
 
 
 class Bundle(SearchStrategy[Ex]):
-    def __init__(self, name: str, consume: bool = False) -> None:
+    """A collection of values for use in stateful testing.
+
+    Bundles are a kind of strategy where values can be added by rules,
+    and (like any strategy) used as inputs to future rules.
+
+    The ``name`` argument they are passed is the they are referred to
+    internally by the state machine; no two bundles may have
+    the same name. It is idiomatic to use the attribute
+    being assigned to as the name of the Bundle::
+
+        class MyStateMachine(RuleBasedStateMachine):
+            keys = Bundle("keys")
+
+    Bundles can contain the same value more than once; this becomes
+    relevant when using :func:`~hypothesis.stateful.consumes` to remove
+    values again.
+
+    If the ``consume`` argument is set to True, then all values that are
+    drawn from this bundle will be consumed (as above) when requested.
+    """
+
+    def __init__(
+        self, name: str, *, consume: bool = False, draw_references: bool = True
+    ) -> None:
         self.name = name
-        self.__reference_strategy = BundleReferenceStrategy(name, consume)
+        self.__reference_strategy = BundleReferenceStrategy(name, consume=consume)
+        self.draw_references = draw_references
 
     def do_draw(self, data):
         machine = data.draw(self_strategy)
@@ -466,7 +537,7 @@ class Bundle(SearchStrategy[Ex]):
         consume = self.__reference_strategy.consume
         if consume is False:
             return f"Bundle(name={self.name!r})"
-        return f"Bundle(name={self.name!r}, consume={consume!r})"
+        return f"Bundle(name={self.name!r}, {consume=})"
 
     def calc_is_empty(self, recur):
         # We assume that a bundle will grow over time
@@ -479,6 +550,15 @@ class Bundle(SearchStrategy[Ex]):
         machine = data.draw(self_strategy)
         return bool(machine.bundle(self.name))
 
+    def flatmap(self, expand):
+        if self.draw_references:
+            return type(self)(
+                self.name,
+                consume=self.__reference_strategy.consume,
+                draw_references=False,
+            ).flatmap(expand)
+        return super().flatmap(expand)
+
 
 class BundleConsumer(Bundle[Ex]):
     def __init__(self, bundle: Bundle[Ex]) -> None:
@@ -529,7 +609,7 @@ def _convert_targets(targets, target):
         if targets:
             raise InvalidArgument(
                 "Passing both targets=%r and target=%r is redundant - pass "
-                "targets=%r instead." % (targets, target, tuple(targets) + (target,))
+                "targets=%r instead." % (targets, target, (*targets, target))
             )
         targets = (target,)
 
@@ -553,6 +633,7 @@ def _convert_targets(targets, target):
                     "This will be an error in a future version of Hypothesis.",
                     since="2021-09-08",
                     has_codemod=False,
+                    stacklevel=2,
                 )
             t = t.name
         converted_targets.append(t)
@@ -578,15 +659,15 @@ _RuleWrapper = Callable[[_RuleType[Ex]], _RuleType[Ex]]
 # a `SearchStrategy`, which the concrete implementation does not accept.
 #
 # Omitted `targets` parameters, where the default value is used, are typed with
-# a special `_OmittedArgument` type. We cannot type them as `Tuple[()]`, because
-# `Tuple[()]` is a subtype of `Sequence[Bundle[Ex]]`, leading to signature
+# a special `_OmittedArgument` type. We cannot type them as `tuple[()]`, because
+# `tuple[()]` is a subtype of `Sequence[Bundle[Ex]]`, leading to signature
 # overlaps with incompatible return types. The `_OmittedArgument` type will never be
 # encountered at runtime, and exists solely to annotate the default of `targets`.
 # PEP 661 (Sentinel Values) might provide a more elegant alternative in the future.
 #
-# We could've also annotated `targets` as `Tuple[_OmittedArgument]`, but then when
+# We could've also annotated `targets` as `tuple[_OmittedArgument]`, but then when
 # both `target` and `targets` are provided, mypy describes the type error as an
-# invalid argument type for `targets` (expected `Tuple[_OmittedArgument]`, got ...).
+# invalid argument type for `targets` (expected `tuple[_OmittedArgument]`, got ...).
 # By annotating it as a bare `_OmittedArgument` type, mypy's error will warn that
 # there is no overloaded signature matching the call, which is more descriptive.
 #
@@ -631,7 +712,7 @@ def rule(
     ``targets`` will define where the end result of this function should go. If
     both are empty then the end result will be discarded.
 
-    ``target`` must be a Bundle, or if the result should go to multiple
+    ``target`` must be a Bundle, or if the result should be replicated to multiple
     bundles you can pass a tuple of them as the ``targets`` argument.
     It is invalid to use both arguments for a single rule.  If the result
     should go to exactly one of several bundles, define a separate rule for
@@ -805,19 +886,19 @@ def precondition(precond: Callable[[Any], bool]) -> Callable[[TestFunc], TestFun
         invariant = getattr(f, INVARIANT_MARKER, None)
         if rule is not None:
             assert invariant is None
-            new_rule = attr.evolve(rule, preconditions=rule.preconditions + (precond,))
+            new_rule = attr.evolve(rule, preconditions=(*rule.preconditions, precond))
             setattr(precondition_wrapper, RULE_MARKER, new_rule)
         elif invariant is not None:
             assert rule is None
             new_invariant = attr.evolve(
-                invariant, preconditions=invariant.preconditions + (precond,)
+                invariant, preconditions=(*invariant.preconditions, precond)
             )
             setattr(precondition_wrapper, INVARIANT_MARKER, new_invariant)
         else:
             setattr(
                 precondition_wrapper,
                 PRECONDITIONS_MARKER,
-                getattr(f, PRECONDITIONS_MARKER, ()) + (precond,),
+                (*getattr(f, PRECONDITIONS_MARKER, ()), precond),
             )
 
         return precondition_wrapper
@@ -892,7 +973,8 @@ class RuleStrategy(SearchStrategy):
         self.rules = list(machine.rules())
 
         self.enabled_rules_strategy = st.shared(
-            FeatureStrategy(), key=("enabled rules", machine)
+            FeatureStrategy(at_least_one_of={r.function.__name__ for r in self.rules}),
+            key=("enabled rules", machine),
         )
 
         # The order is a bit arbitrary. Primarily we're trying to group rules
@@ -911,10 +993,7 @@ class RuleStrategy(SearchStrategy):
         )
 
     def __repr__(self):
-        return "{}(machine={}({{...}}))".format(
-            self.__class__.__name__,
-            self.machine.__class__.__name__,
-        )
+        return f"{self.__class__.__name__}(machine={self.machine.__class__.__name__}({{...}}))"
 
     def do_draw(self, data):
         if not any(self.is_valid(rule) for rule in self.rules):
@@ -923,26 +1002,39 @@ class RuleStrategy(SearchStrategy):
 
         feature_flags = data.draw(self.enabled_rules_strategy)
 
-        # Note: The order of the filters here is actually quite important,
-        # because checking is_enabled makes choices, so increases the size of
-        # the choice sequence. This means that if we are in a case where many
-        # rules are invalid we will make a lot more choices if we ask if they
-        # are enabled before we ask if they are valid, so our test cases will
-        # be artificially large.
-        rule = data.draw(
-            st.sampled_from(self.rules)
-            .filter(self.is_valid)
-            .filter(lambda r: feature_flags.is_enabled(r.function.__name__))
-        )
+        def rule_is_enabled(r):
+            # Note: The order of the filters here is actually quite important,
+            # because checking is_enabled makes choices, so increases the size of
+            # the choice sequence. This means that if we are in a case where many
+            # rules are invalid we would make a lot more choices if we ask if they
+            # are enabled before we ask if they are valid, so our test cases would
+            # be artificially large.
+            return self.is_valid(r) and feature_flags.is_enabled(r.function.__name__)
 
-        return (rule, data.draw(rule.arguments_strategy))
+        rule = data.draw(st.sampled_from(self.rules).filter(rule_is_enabled))
+
+        arguments = {}
+        for k, strat in rule.arguments_strategies.items():
+            try:
+                arguments[k] = data.draw(strat)
+            except Exception as err:
+                rname = rule.function.__name__
+                add_note(err, f"while generating {k!r} from {strat!r} for rule {rname}")
+                raise
+        return (rule, arguments)
 
     def is_valid(self, rule):
-        if not all(precond(self.machine) for precond in rule.preconditions):
-            return False
-
         for b in rule.bundles:
-            bundle = self.machine.bundle(b.name)
-            if not bundle:
+            if not self.machine.bundle(b.name):
                 return False
+
+        predicates = self.machine._observability_predicates
+        desc = f"{self.machine.__class__.__qualname__}, rule {rule.function.__name__},"
+        for pred in rule.preconditions:
+            meets_precond = pred(self.machine)
+            where = f"{desc} precondition {get_pretty_function_description(pred)}"
+            predicates[where]["satisfied" if meets_precond else "unsatisfied"] += 1
+            if not meets_precond:
+                return False
+
         return True
